"""`PySpelling` package."""

#   (major, minor, micro, release type, pre-release build, post-release build, development-release)
version_info = (1, 0, 0, 'beta', 1, 0)


def pep440_version(vi):
    """
    Get the version (PEP 440).

    Version structure
      (major, minor, micro, release type, pre-release build, post-release build)
    Release names are named is such a way they are sortable and comparable with ease.
      (alpha | beta | candidate | final)

    - "final" should never have a pre-release build number
    - pre-releases should have a pre-release build number greater than 0
    - post-release is only applied if post-release build is greater than 0
    - development-release is only applied if `.dev-` is appended to release type.
      It can be applied to pre and post releases. Intention is to only use this internally
      and never actually release a `dev` to the public server. As we don't manage build releases,
      we will just use `dev0`.

    Acceptable version releases:

    ~~~
    (1, 0, 0, 'final', 0, 0)      1.0
    (1, 2, 0, 'final', 0, 0)      1.2
    (1, 2, 3, 'final', 0, 0)      1.2.3
    (1, 2, 0, 'alpha', 4, 0):     1.2a4
    (1, 2, 0, 'beta', 4, 0):      1.2b4
    (1, 2, 0, 'candidate', 4, 0): 1.2rc4
    (1, 2, 0, 'final', 0, 1)      1.2.post1
    (1, 2, 0, 'beta', 1, 1):      1.2b1.post1
    (1, 2, 3, '.dev-final', 0, 0) 1.2.3.dev0
    (1, 2, 3, '.dev-alpha', 1, 0) 1.2.3a1.dev0
    (1, 2, 3, '.dev-final', 0, 1) 1.2.3.post1.dev0
    (1, 2, 3, '.dev-beta', 2, 1)  1.2.3b2.post1.dev0
    ~~~

    """

    # Determine if this is a development build or regular build
    dev, rel_type = (True, vi[3][5:]) if vi[3].startswith('.dev-') else (False, vi[3])

    releases = {"alpha": 'a', "beta": 'b', "candidate": 'rc', "final": ''}

    # Ensure version info is valid.
    # Version info should be proper length.
    assert len(vi) == 6
    # Should be a valid release.
    assert rel_type in releases
    # Pre-release releases should have a pre-release value.
    assert rel_type == 'final' or vi[4] > 0
    # Final should not have a pre-release value.
    assert rel_type != 'final' or vi[4] == 0

    # Assemble major, minor, patch version. Patch can be left out if 0.
    main = '.'.join(str(x) for x in (vi[0:2] if vi[2] == 0 else vi[0:3]))
    # Assemble pre-release if needed.
    prerel = releases[rel_type] + str(vi[4]) if releases[rel_type] else ''
    # Assemble post-release if we have one.
    postrel = '.post%d' % vi[5] if vi[5] > 0 else ''
    # Assemble development release if we have one.
    devrel = '.dev0' if dev else ''

    return ''.join((main, prerel, postrel, devrel))


version = pep440_version(version_info)



"""Stylesheets filter."""
import re
from . import cpp

RE_CSS_COMMENT = re.compile(
    r'''(?x)
        (?P<comments>
            (?P<block>/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)                        # multi-line comments
        )
      | (?P<code>
            "(?:\\.|[^"\\])*"                                                # double quotes
            '(?:\\.|[^'\\])*'                                                # single quotes
          | .[^/"']*?                                                        # everything else
        )
    ''',
    re.DOTALL | re.MULTILINE
)

CSS = 0
SASS = 1
SCSS = 2

STYLESHEET_TYPE = {
    "css": CSS,
    "sass": SASS,
    "scss": SCSS
}


class StylesheetsFilter(cpp.CppFilter):
    """Stylesheets filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        # If the style isn't found, just go with CSS, then use the appropriate prefix.
        self.stylesheets = STYLESHEET_TYPE.get(options.get('stylesheets', 'css').lower(), CSS)
        self.prefix = [k for k, v in STYLESHEET_TYPE.items() if v == SASS][0]
        super(StylesheetsFilter, self).__init__(options, default_encoding)

    def find_comments(self, text):
        """Find comments."""

        if self.stylesheets == CSS:
            for m in RE_CSS_COMMENT.finditer(text):
                self._css_evaluate(m)
        else:
            super(StylesheetsFilter, self).find_comments(text)

    def _css_evaluate(self, m):
        """Search for comments."""

        g = m.groupdict()
        if g["code"]:
            self.line_num += g["code"].count('\n')
        else:
            self.evaluate_block(g)
            self.line_num += g['comments'].count('\n')

    def extend_src(self, content, context, encoding):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, encoding, 'block-comment')
        if self.stylesheets != CSS:
            self.extend_src_text(content, context, self.line_comments, encoding, 'line-comment')


def get_plugin():
    """Get filter."""

    return StylesheetsFilter



"""
HTML filter.

Detect encoding from HTML header.
"""
from __future__ import unicode_literals
from .. import filters
import re
import codecs
import bs4
from collections import namedtuple
from html.parser import HTMLParser

RE_XML_START = re.compile(
    b'^(?:(' +
    b'<\\?xml[^>]+?>' +  # ASCII like
    b')|(' +
    re.escape('<?xml'.encode('utf-32-be')) + b'.+?' + re.escape('>'.encode('utf-32-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-32-le')) + b'.+?' + re.escape('>'.encode('utf-32-le')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-be')) + b'.+?' + re.escape('>'.encode('utf-16-be')) +
    b')|(' +
    re.escape('<?xml'.encode('utf-16-le')) + b'.+?' + re.escape('>'.encode('utf-16-le')) +
    b'))'
)

RE_HTML_ENCODE = re.compile(
    br'''(?xi)
    <\s*meta(?!\s*(?:name|value)\s*=)(?:[^>]*?content\s*=[\s"']*)?(?:[^>]*?)[\s"';]*charset\s*=[\s"']*([^\s"'/>]*)
    '''
)

RE_XML_ENCODE = re.compile(br'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')
RE_XML_ENCODE_U = re.compile(r'''(?i)^<\?xml[^>]*encoding=(['"])(.*?)\1[^>]*\?>''')

RE_SELECTOR = re.compile(r'''(\#|\.)?[-\w]+|\*|\[([\w\-:]+)(?:([~^|*$]?=)(\"[^"]+\"|'[^']'|[^'"\[\]]+))?\]''')


class Selector(namedtuple('IgnoreRule', ['tag', 'id', 'classes', 'attributes'])):
    """Ignore rule."""


class SelectorAttribute(namedtuple('AttrRule', ['attribute', 'pattern'])):
    """Selector attribute rule."""


class HtmlFilter(filters.Filter):
    """Spelling Python."""

    block_tags = [
        # Block level elements (and other blockish elements)
        'address', 'article', 'aside', 'blockquote', 'details', 'dialog', 'dd',
        'div', 'dl', 'dt'
        'fieldset', 'figcaption', 'figure', 'footer', 'form', 'h1', 'h2', 'h3',
        'h4', 'h5', 'h6', 'header', 'hgroup', 'hr', 'li', 'main', 'menu', 'nav', 'ol', 'p', 'pre',
        'section', 'table', 'ul',
        'canvas', 'group', 'iframe', 'math', 'noscript', 'output',
        'script', 'style', 'table', 'video', 'body', 'head'
    ]

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.html_parser = HTMLParser()
        self.comments = options.get('comments', True) is True
        self.attributes = set(options.get('attributes', []))
        self.selectors = self.process_selectors(*options.get('ignores', []))
        super(HtmlFilter, self).__init__(options, default_encoding)

    def _has_xml_encode(self, content):
        """Check XML encoding."""

        encode = None

        m = RE_XML_START.match(content)
        if m:
            if m.group(1):
                m2 = RE_XML_ENCODE.match(m.group(1))

                if m2:
                    enc = m2.group(2).decode('ascii')

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except LookupError:
                        pass
            else:
                if m.group(2):
                    enc = 'utf-32-be'
                    text = m.group(2)
                elif m.group(3):
                    enc = 'utf-32-le'
                    text = m.group(3)
                elif m.group(4):
                    enc = 'utf-16-be'
                    text = m.group(4)
                elif m.group(5):
                    enc = 'utf-16-le'
                    text = m.group(5)
                try:
                    m2 = RE_XML_ENCODE_U.match(text.decode(enc))
                except Exception:  # pragma: no cover
                    m2 = None

                if m2:
                    enc = m2.group(2)

                    try:
                        codecs.getencoder(enc)
                        encode = enc
                    except Exception:
                        pass

        return encode

    def header_check(self, content):
        """Special HTML encoding check."""

        encode = None

        # Look for meta charset
        m = RE_HTML_ENCODE.search(content)
        if m:
            enc = m.group(1).decode('ascii')

            try:
                codecs.getencoder(enc)
                encode = enc
            except LookupError:
                pass
        else:
            encode = self._has_xml_encode(content)
        return encode

    def is_block(self, el):
        """Check if tag is a block element."""

        return el.name.lower() in self.block_tags

    def process_selectors(self, *args):
        """
        Process selectors.

        We do our own selectors as BeautifulSoup4 has some annoying quirks,
        and we don't really need to do nth selectors or siblings or
        descendants etc.
        """

        selectors = [
            Selector('style', None, tuple(), tuple()),
            Selector('script', None, tuple(), tuple()),
        ]

        for selector in args:
            tag = None
            tag_id = None
            classes = set()
            attributes = []

            for m in RE_SELECTOR.finditer(selector):
                if m.group(2):
                    attr = m.group(2).lower()
                    op = m.group(3)
                    if op:
                        value = m.group(4)[1:-1] if m.group(4).startswith('"') else m.group(4)
                    else:
                        value = None
                    if not op:
                        # Attribute name
                        pattern = None
                    elif op.startswith('^'):
                        # Value start with
                        pattern = re.compile(r'^%s.*' % re.escape(value))
                    elif op.startswith('$'):
                        # Value ends with
                        pattern = re.compile(r'.*?%s$' % re.escape(value))
                    elif op.startswith('*'):
                        # Value contains
                        pattern = re.compile(r'.*?%s.*' % re.escape(value))
                    elif op.startswith('~'):
                        # Value contains word within space separated list
                        pattern = re.compile(r'.*?(?:(?<=^)|(?<= ))%s(?=(?:[ ]|$)).*' % re.escape(value))
                    elif op.startswith('|'):
                        # Value starts with word in dash separated list
                        pattern = re.compile(r'^%s(?=-).*' % re.escape(value))
                    else:
                        # Value matches
                        pattern = re.compile(r'^%s$' % re.escape(value))
                    attributes.append(SelectorAttribute(attr, pattern))
                else:
                    selector = m.group(0).lower()
                    if selector.startswith('.'):
                        classes.add(selector[1:].lower())
                    elif selector.startswith('#') and tag_id is None:
                        tag_id = selector[1:]
                    elif tag is None:
                        tag = selector
                    else:
                        raise ValueError('Bad selector!')

            if tag or tag_id or classes:
                selectors.append(Selector(tag, tag_id, tuple(classes), tuple(attributes)))

        return selectors

    def skip_tag(self, el):
        """Determine if tag should be skipped."""

        skip = False
        for selector in self.selectors:
            if selector.tag and selector.tag not in (el.name.lower(), '*'):
                continue
            if selector.id and selector.id != el.attrs.get('id', '').lower():
                continue
            if selector.classes:
                current_classes = [c.lower() for c in el.attrs.get('class', [])]
                found = True
                for c in selector.classes:
                    if c not in current_classes:
                        found = False
                        break
                if not found:
                    continue
            if selector.attributes:
                found = True
                for a in selector.attributes:
                    value = el.attrs.get(a.attribute)
                    if isinstance(value, list):
                        value = ' '.join(value)
                    if not value:
                        found = False
                        break
                    elif a.pattern is None:
                        continue
                    elif a.pattern.match(value) is None:
                        found = False
                        break
                if not found:
                    continue
            skip = True
            break
        return skip

    def construct_selector(self, el, attr=''):
        """Construct an selector for context."""

        tag = el.name
        if tag == '[document]':
            tag = 'HTML'
        classes = el.attrs.get('class', [])
        tag_id = el.attrs.get('id', '').strip()
        sel = tag.upper()
        if tag_id:
            sel += '#' + tag_id
        if classes:
            sel += '.' + '.'.join(classes)
        if attr:
            sel += '[%s]' % attr
        return sel

    def html_to_text(self, tree):
        """
        Parse the HTML creating a buffer with each tags content.

        Skip any selectors specified and include attributes if specified.
        Ignored tags will not have their attributes scanned either.
        """

        text = []
        attributes = []
        comments = []
        blocks = []
        root = tree.name == '[document]'

        # Handle comments
        if isinstance(tree, bs4.Comment):
            if self.comments:
                string = str(tree).strip()
                if string:
                    comments.append(self.html_parser.unescape(string))
        elif root or not self.skip_tag(tree):
            # Check attributes for normal tags
            if not root:
                for attr in self.attributes:
                    value = tree.attrs.get(attr, '').strip()
                    if value:
                        sel = self.construct_selector(tree, attr=attr)
                        attributes.append((self.html_parser.unescape(value), sel))

            # Walk children
            for child in tree:
                if isinstance(child, bs4.element.Tag):
                    if child.contents:
                        t, b, a, c = (self.html_to_text(child))
                        text.extend(t)
                        attributes.extend(a)
                        comments.extend(c)
                        blocks.extend(b)
                # Get content if not the root and not a comment (unless we want comments).
                elif not root and (not isinstance(child, bs4.Comment) or self.comments):
                    string = str(child).strip()
                    if string:
                        text.append(string)
                        text.append(' ')

        if root or self.is_block(tree):
            content = self.html_parser.unescape(''.join(text))
            if content:
                blocks.append((content, self.construct_selector(tree)))
            text = []

        if root:
            return (
                blocks,
                attributes,
                comments
            )
        else:
            return text, blocks, attributes, comments

    def _filter(self, text, context, encoding):
        """Filter the source text."""

        content = []
        blocks, attributes, comments = self.html_to_text(bs4.BeautifulSoup(text, "html5lib"))
        if self.comments:
            for c in comments:
                content.append(filters.SourceText(c, context + ': <!-- comment -->', encoding, 'html-comment'))
        if self.attributes:
            for a, desc in attributes:
                content.append(filters.SourceText(a, context + ': ' + desc, encoding, 'html-attribute'))
        for b, desc in blocks:
            content.append(filters.SourceText(b, context + ': ' + desc, encoding, 'html-content'))
        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse HTML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Return the filter."""

    return HtmlFilter



"""Main."""
from __future__ import unicode_literals
import sys
import argparse
from pyspelling import spellcheck, __version__


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='spellcheck', description='Spell checking tool.')
    # Flag arguments
    parser.add_argument('--version', action='version', version=('%(prog)s ' + __version__))
    parser.add_argument('--debug', action='store_true', default=False, help=argparse.SUPPRESS)
    parser.add_argument('--verbose', '-v', action='count', default=0, help="Verbosity level.")
    parser.add_argument('--name', '-n', action='store', default='', help="Specific spelling task by name to run.")
    parser.add_argument('--binary', '-b', action='store', default='', help="Provide path to spell checker's binary.")
    parser.add_argument('--config', '-c', action='store', default='', help="Spelling config.")
    parser.add_argument(
        '--spellchecker', '-s', action='store', default='', help="Choose between aspell and hunspell"
    )
    args = parser.parse_args()

    return run(args.config, args.name, args.binary, args.verbose, args.spellchecker, args.debug)


def run(config, name='', binary='', verbose=0, spellchecker='', debug=False):
    """Run."""

    fail = False
    count = 0
    for results in spellcheck(config, name, binary, verbose, spellchecker, debug):
        count += 1
        if results.error:
            fail = True
            print('ERROR: %s -- %s' % (results.context, results.error))
        elif results.words:
            fail = True
            print('Misspelled words:\n<%s> %s' % (results.category, results.context))
            print('-' * 80)
            for word in results.words:
                print(word)
            print('-' * 80)
            print('')

    if fail:
        print('!!!Spelling check failed!!!')
    else:
        print('Spelling check passed :)')

    return fail


if __name__ == "__main__":
    sys.exit(main())



"""JavaScript filter."""
import re
from . import cpp

RE_JSDOC = re.compile(r"(?s)^/\*\*$(.*?)[ \t]*\*/", re.MULTILINE)


class JavaScriptFilter(cpp.CppFilter):
    """JavaScript filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.jsdocs = options.get('jsdocs', False) is True
        options['prefix'] = 'js'
        super(JavaScriptFilter, self).__init__(options, default_encoding)

    def evaluate_block(self, groups):
        """Evaluate block comments."""

        if self.jsdocs:
            m1 = RE_JSDOC.match(groups['comments'])
            if m1:
                lines = []
                for line in m1.group(1).splitlines(True):
                    l = line.lstrip()
                    lines.append(l[1:] if l.startswith('*') else l)
                self.jsdoc_comments.append([''.join(lines), self.line_num])
            elif self.blocks:
                self.block_comments.append([groups['block'][2:-2], self.line_num])
        elif self.blocks:
            self.block_comments.append([groups['block'][2:-2], self.line_num])

    def extend_src(self, content, context, encoding):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, encoding, 'block-comment')
        self.extend_src_text(content, context, self.line_comments, encoding, 'line-comment')
        self.extend_src_text(content, context, self.jsdoc_comments, encoding, 'docs')

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        self.jsdoc_comments = []
        return super(JavaScriptFilter, self)._filter(text, context, encoding)


def get_plugin():
    """Get filter."""

    return JavaScriptFilter



"""Filter Plugins."""
from __future__ import unicode_literals
import re
import codecs
import contextlib
import mmap
import os
from collections import namedtuple

PYTHON_ENCODING_NAMES = {
    'iso-8859-8-i': 'iso-8859-8',
    'macintosh': 'mac-roman',
    'windows-874': 'cp874',
    'x-mac-cyrillic': 'mac-cyrillic',
    "x-sjis": "shift-jis"
}

RE_UTF_BOM = re.compile(
    b'^(?:(' +
    codecs.BOM_UTF8 +
    b')[\x00-\xFF]{,2}|(' +
    codecs.BOM_UTF32_BE +
    b')|(' +
    codecs.BOM_UTF32_LE +
    b')|(' +
    codecs.BOM_UTF16_BE +
    b')|(' +
    codecs.BOM_UTF16_LE +
    b'))'
)

RE_CATEGORY_NAME = re.compile(r'^[-a-z0-9_]+$', re.I)


class SourceText(namedtuple('SourceText', ['text', 'context', 'encoding', 'category', 'error'])):
    """Source text."""

    __slots__ = ()

    def __new__(cls, text, context, encoding, category, error=None):
        """Allow defaults."""

        encoding = PYTHON_ENCODING_NAMES.get(encoding, encoding).lower()

        if encoding == 'utf-8-sig':
            encoding = 'utf-8'
        if encoding.startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.startswith('utf-32'):
            encoding = 'utf-32'

        if encoding:
            encoding = codecs.lookup(encoding).name

        if RE_CATEGORY_NAME.match(category) is None and error is None:
            raise ValueError('Invalid category name in SourceText!')

        return super(SourceText, cls).__new__(cls, text, context, encoding, category, error)

    def _is_bytes(self):
        """Is bytes."""

        return isinstance(self.text, bytes)

    def _has_error(self):
        """Check if object has an error associated with it."""

        return self.error is not None


class Filter(object):
    """Spelling language."""

    MAX_GUESS_SIZE = 31457280
    CHECK_BOM = True

    def __init__(self, config, default_encoding='utf-8'):
        """Initialize."""

        self.config = config
        self.default_encoding = PYTHON_ENCODING_NAMES.get(default_encoding, default_encoding).lower()

    def _is_very_large(self, size):
        """Check if content is very large."""

        return size >= self.MAX_GUESS_SIZE

    def _verify_encoding(self, enc):
        """Verify encoding is okay."""

        enc = PYTHON_ENCODING_NAMES.get(enc, enc)
        try:
            codecs.getencoder(enc)
            encoding = enc
        except LookupError:
            encoding = None
        return encoding

    def _has_bom(self, content):
        """Check for UTF8, UTF16, and UTF32 BOMs."""

        encoding = None
        m = RE_UTF_BOM.match(content)
        if m is not None:
            if m.group(1):
                encoding = 'utf-8-sig'
            elif m.group(2):
                encoding = 'utf-32'
            elif m.group(3):
                encoding = 'utf-32'
            elif m.group(4):
                encoding = 'utf-16'
            elif m.group(5):
                encoding = 'utf-16'
        return encoding

    def _utf_strip_bom(self, encoding):
        """Return an encoding that will ignore the BOM."""

        if encoding is None:
            pass
        elif encoding.lower() == 'utf-8':
            encoding = 'utf-8-sig'
        elif encoding.lower().startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.lower().startswith('utf-32'):
            encoding = 'utf-32'
        return encoding

    def _detect_buffer_encoding(self, f):
        """Guess by checking BOM, and checking `_special_encode_check`, and using memory map."""

        encoding = None
        with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
            # Check for BOMs
            if self.CHECK_BOM:
                encoding = self._has_bom(m.read(4))
            m.seek(0)
            # Check file extensions
            if encoding is None:
                encoding = self._utf_strip_bom(self.header_check(m.read(1024)))
                m.seek(0)
            if encoding is None:
                m.seek(0)
                encoding = self._utf_strip_bom(self.content_check(m))

        return encoding

    def _detect_encoding(self, source_file):
        """Detect encoding."""

        encoding = self._guess(source_file)
        # If we didn't explicitly detect an encoding, assume default.
        if not encoding:
            encoding = self.default_encoding

        return encoding

    def _parse(self, source_file):
        """Parse the file."""

        self.current_encoding = self.default_encoding
        encoding = None
        try:
            encoding = self._detect_encoding(source_file)
            content = self.filter(source_file, encoding)
        except UnicodeDecodeError as e:
            if not encoding or encoding != self.default_encoding:
                content = self.filter(source_file, self.default_encoding)
            else:
                raise
        return content

    def _guess(self, filename):
        """Guess the encoding and decode the content of the file."""

        encoding = None

        try:
            file_size = os.path.getsize(filename)
            # If the file is really big, lets just call it binary.
            # We don't have time to let Python chug through a massive file.
            if not self._is_very_large(file_size):
                with open(filename, "rb") as f:
                    if file_size == 0:
                        encoding = 'ascii'
                    else:
                        encoding = self._detect_buffer_encoding(f)
                        if encoding is not None:
                            encoding = self._verify_encoding(encoding)
            else:
                raise UnicodeDecodeError('Unicode detection is not applied to very large files!')
        except Exception:  # pragma: no cover
            raise UnicodeDecodeError('Cannot resolve encoding!')
            pass

        return encoding

    def content_check(self, file_handle):
        """File content check."""

        return None

    def header_check(self, content):
        """Special encode check."""

        return None

    def filter(self, source_file, encoding):  # noqa A001
        """Open and filter the file from disk."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return [SourceText(text, source_file, encoding, 'text')]

    def sfilter(self, source):
        """Execute filter."""

        return [SourceText(source.text, source.context, source.encoding, 'text')]



"""Flow control."""

ALLOW = 0
SKIP = 1
HALT = 2


class FlowControl:
    """Control flow of objects in the pipeline."""

    def __init__(self, config):
        """Initialization."""

    def adjust_flow(self, category):
        """Adjust the flow of source control objects."""

        return ALLOW



"""Text parser."""
from __future__ import unicode_literals
from .. import filters
from .. import util
import codecs
import re

DEFAULT_CONTENT = '.*?'


class ContextFilter(filters.Filter):
    """Context filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.context_visible_first = options.get('context_visible_first', False) is True
        self.delimiters = []
        self.escapes = None
        escapes = []
        for delimiter in options.get('delimiters', []):
            if not isinstance(delimiter, dict):
                continue
            group = util.random_name_gen()
            while (
                group in delimiter['open'] or
                group in delimiter['close'] or
                group in delimiter.get('content', DEFAULT_CONTENT)
            ):
                group = util.random_name_gen()

            pattern = r'%s(?P<%s>%s)(?:%s|\Z)' % (
                delimiter['open'],
                group,
                delimiter.get('content', DEFAULT_CONTENT),
                delimiter['close']
            )
            self.delimiters.append((re.compile(pattern, re.M), group))
        escapes = options.get('escapes', None)
        if escapes:
            self.escapes = re.compile(escapes)

        super(ContextFilter, self).__init__(options, default_encoding)

    def filter(self, source_file, encoding):  # noqa A001
        """Parse file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()

        return [filters.SourceText(self._filter(text), source_file, encoding, 'context')]

    def sfilter(self, source):
        """Filter."""

        return [filters.SourceText(self._filter(source.text), source.context, source.encoding, 'context')]

    def _filter(self, text):
        """Context delimiter filter."""

        new_text = []
        index = 0
        last = 0
        end = len(text)
        while index < end:
            m = self.escapes.match(text, pos=index) if self.escapes else None
            if m:
                index = m.end(0)
                continue
            handled = False
            for delimiter in self.delimiters:
                m = delimiter[0].match(text, pos=index)
                if m:
                    if self.context_visible_first is True:
                        new_text.append(text[last:m.start(0)])
                    else:
                        new_text.append(m.group(delimiter[1]))
                    index = m.end(0)
                    last = index
                    handled = True
                    break
            if handled:
                continue
            index += 1
        if last < end and self.context_visible_first is True:
            new_text.append(text[last:end])

        return ' '.join(new_text)


def get_plugin():
    """Return the filter."""

    return ContextFilter



"""Text filter."""
from __future__ import unicode_literals
from .. import filters
import codecs
import unicodedata


class TextFilter(filters.Filter):
    """Spelling Text."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.normalize = options.get('normalize', '').upper()
        self.convert_encoding = options.get('convert_encoding', '').lower()
        self.errors = options.get('errors', 'strict').lower()
        if self.convert_encoding:
            self.convert_encoding = codecs.lookup(
                filters.PYTHON_ENCODING_NAMES.get(default_encoding, default_encoding).lower()
            ).name

            # Don't generate content with BOMs
            if (
                self.convert_encoding.startswith(('utf-32', 'utf-16')) and
                not self.convert_encoding.endswith(('le', 'be'))
            ):
                self.convert_encoding += '-le'

            if self.convert_encoding == 'utf-8-sig':
                self.convert_encoding = 'utf-8'

        super(TextFilter, self).__init__(options, default_encoding)

    def convert(self, text, encoding):
        """Convert the text."""

        if self.normalize in ('NFC', 'NFKC', 'NFD', 'NFKD'):
            text = unicodedata.normalize(self.normalize, text)
        if self.convert_encoding:
            text = text.encode(self.convert_encoding, self.errors).decode(self.convert_encoding)
            encoding = self.convert_encoding
        return text, encoding

    def filter(self, source_file, encoding):  # noqa A001
        """Open and filter the file from disk."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()

        text, encoding = self.convert(text, encoding)

        return [filters.SourceText(text, source_file, encoding, 'text')]

    def sfilter(self, source):
        """Execute filter."""

        text, encoding = self.convert(source.text, source.encoding)

        return [filters.SourceText(text, source.context, encoding, 'text')]


def get_plugin():
    """Return the filter."""

    return TextFilter



"""Wildcard flow control."""
from .. import flow_control
from wcmatch import fnmatch


class WildcardFlowControl(flow_control.FlowControl):
    """Control flow of objects in the pipeline with wildcard patterns."""

    FNMATCH_FLAGS = fnmatch.N | fnmatch.B | fnmatch.I

    def __init__(self, config):
        """Initialization."""

        self.allow = config.get('allow', ['*'])
        self.halt = config.get('halt', [])
        self.skip = config.get('skip', [])

    def match(self, category, pattern):
        """Match the category."""

        return fnmatch.fnmatch(category, fnmatch.fnsplit(pattern, flags=self.FNMATCH_FLAGS), flags=self.FNMATCH_FLAGS)

    def adjust_flow(self, category):
        """Adjust the flow of source control objects."""

        status = flow_control.SKIP
        for allow in self.allow:
            if self.match(category, allow):
                status = flow_control.ALLOW
                for skip in self.skip:
                    if self.match(category, skip):
                        status = flow_control.SKIP
                for halt in self.halt:
                    if self.match(category, halt):
                        status = flow_control.HALT
                if status != flow_control.ALLOW:
                    break
        return status


def get_plugin():
    """Get flow controller."""

    return WildcardFlowControl



"""C++ filter."""
import re
import codecs
from .. import filters
import textwrap

RE_COMMENT = re.compile(
    r'''(?x)
        (?P<comments>
            (?P<block>/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)                        # multi-line comments
          | (?P<start>^)?(?P<leading_space>[ \t]*)?(?P<line>//(?:[^\r\n])*)  # single line comments
        )
      | (?P<code>
            "(?:\\.|[^"\\])*"                                                # double quotes
            '(?:\\.|[^'\\])*'                                                # single quotes
          | .[^/"']*?                                                        # everything else
        )
    ''',
    re.DOTALL | re.MULTILINE
)


class CppFilter(filters.Filter):
    """C++ style comment filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.blocks = options.get('block_comments', True) is True
        self.lines = options.get('line_comments', True) is True
        self.group_comments = options.get('group_comments', False) is True
        self.prefix = options.get('prefix', 'cpp')
        super(CppFilter, self).__init__(options, default_encoding)

    def evaluate_block(self, groups):
        """Evaluate block comments."""

        if self.blocks:
            self.block_comments.append([groups['block'][2:-2], self.line_num])

    def evaluate_inline_tail(self, groups):
        """Evaluate inline comments at the tail of source code."""

        if self.lines:
            self.line_comments.append([groups['line'][2:], self.line_num])

    def evaluate_inline(self, groups):
        """Evaluate inline comments on their own lines."""

        # Consecutive lines with only comments with same leading whitespace
        # will be captured as a single block.
        if self.lines:
            if (
                self.group_comments and
                self.line_num == self.prev_line + 1 and
                groups['leading_space'] == self.leading
            ):
                self.line_comments[-1][0] += '\n' + groups['line'][2:]
            else:
                self.line_comments.append([groups['line'][2:], self.line_num])
            self.leading = groups['leading_space']
            self.prev_line = self.line_num

    def _evaluate(self, m):
        """Search for comments."""

        g = m.groupdict()
        if g["code"]:
            self.line_num += g["code"].count('\n')
        else:
            if g['block']:
                self.evaluate_block(g)
            elif g['start'] is None:
                self.evaluate_inline_tail(g)
            else:
                self.evaluate_inline(g)
            self.line_num += g['comments'].count('\n')

    def extend_src_text(self, content, context, text_list, encoding, category):
        """Extend the source text list with the gathered text data."""

        prefix = self.prefix + '-' if self.prefix else ''

        for comment, line in text_list:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    prefix + category
                )
            )

    def extend_src(self, content, context, encoding):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, encoding, 'block-comment')
        self.extend_src_text(content, context, self.line_comments, encoding, 'line-comment')

    def find_comments(self, text):
        """Find comments."""

        for m in RE_COMMENT.finditer(text):
            self._evaluate(m)

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        content = []
        self.line_num = 1
        self.prev_line = -1
        self.leading = ''
        self.block_comments = []
        self.line_comments = []

        self.find_comments(text)
        self.extend_src(content, context, encoding)

        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse HTML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()

        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Get filter."""

    return CppFilter
